// ASPI for Win32 [WNASPI32.DLL] ֘A[` 

#include <windows.h>
#include "aspi32.h"

extern "C" __declspec(dllimport) unsigned long __cdecl GetASPI32SupportInfo(void);
extern "C" __declspec(dllimport) unsigned long __cdecl SendASPI32Command(void*);

enum {	// ASPI Command Code
	SC_HA_INQUIRY,			//00
	SC_GET_DEV_TYPE,		//01
	SC_EXEC_SCSI_CMD,		//02
	SC_SRB_ABORT_SRB,		//03
	SC_RESET_DEV, 			//04
	SC_GET_DISK_INFO = 6,	//06
	SC_RESCAN_SCSI_BUS,		//07
};

enum {	// ASPI Command Status
	SS_PENDING,				//0x00
	SS_COMPLETE,			//0x01
	SS_ABORTED,				//0x02
	SS_ABORT_FAIL,			//0x03
	SS_ERR,					//0x04
	SS_INVALID_REQ = 0x80,	//0x80
	SS_INVALID_HA,			//0x81
	SS_NO_DEVICE,			//0x82
	SS_INVALID_SRB = 0xe0,	//0xe0
	SS_BUFFER_ALIGN,		//0xe1
	SS_FAILED_INIT = 0xe4,	//0xe4
	SS_ASPI_IS_BUSY,		//0xe5
	SS_BUFFER_TO_BIG,		//0xe6
};

// ASPI Request Flags
enum {
	SRB_POSTING				=	0x01,
	SRB_ENABLE_RESIDUAL_COUNT =	0x04,
	SRB_DIR_IN				=	0x08,
	SRB_DIR_OUT				=	0x10,
	SRB_EVENT_NOTIFY		=	0x40,
};

// SCSI Command Messages
#define	SCSI_TestUnitReady		0x00
#define	SCSI_RequestSense		0x03
#define	SCSI_Inquiry			0x12
#define	SCSI_ModeSelect			0x15
#define	SCSI_Copy				0x18
#define	SCSI_ModeSense			0x1a
#define SCSI_StartStopUnit		0x1b
#define	SCSI_ReceiveDiagnostic	0x1c
#define	SCSI_SendDiagnostic		0x1d
#define	SCSI_Compare			0x39
#define	SCSI_CopyAndVerify		0x3a
#define	SCSI_WriteBuffer		0x3b
#define	SCSI_ReadBuffer			0x3c
#define	SCSI_ChangeDefinition	0x40
#define	SCSI_LogSelect			0x4c
#define	SCSI_LogSense			0x4d

// SCSI Target Status
#define	TS_NO			0x00
#define	TS_CHECK		0x02
#define	TS_BUSY			0x08
#define	TS_CONFLICT		0x18

#pragma pack(1)
struct SRB_HAinq
{
	BYTE	CommandCode;
	BYTE	CommandStatus;
	BYTE	HostAdapterID;
	BYTE	RequestFlags;
	DWORD	Reserved;	// Should be 0.
	BYTE	NumberOfHostAdapters;
	BYTE	HostAdapterSCSIID;
	BYTE	ManagerID [16];
	BYTE	AdapterID [16];
	BYTE	Parameters [16];
	WORD	Reserved2[2];	// Should be 0.
};

struct SRB_ExecIO
{
	BYTE	CommandCode;
	BYTE	CommandStatus;
	BYTE	HostAdapterID;
	BYTE	RequestFlags;
	DWORD	Reserved;	// Should be 0.
	BYTE	TargetID;
	BYTE	TargetLUN;
	WORD	Reserved1;	// Should be 0.
	DWORD	DataAllocationLength;
	PBYTE	DataBufferPointer;
	BYTE	SenseAllocationLength;
	BYTE	CDBLength;
	BYTE	HostAdapterStatus;
	BYTE	TargetStatus;
	void*	PostProc;
	BYTE	Reserved2[20];	// Should be 0.
	BYTE	CDB[16];
	BYTE	SenseArea[16];
};
#pragma pack()

int OptimizeList();
int AddListItem( HANDLE hThread );
BOOL PerformAspiCmd( SRB_ExecIO* );
BOOL PerformAspiCmdThread( SRB_ExecIO* psrb );
BOOL PerformAspiCmdPending( SRB_ExecIO* psrb );

// CXg[`FbN
int CheckInstallAspi32(void)
{
	// WinASPI32 CXg[Ă邩ׂ
	int NumberOfHA = GetASPI32SupportInfo();
	if ( ( NumberOfHA & 0xff00 ) != 0x0100 ) NumberOfHA = 0;
	else NumberOfHA &= 0xff;
	return NumberOfHA;
}

// ASPI֐Q
int HA_Inquiry( int adapter )
{
	SRB_HAinq	srb;
	ZeroMemory( &srb, sizeof(srb) );
	srb.CommandCode = SC_HA_INQUIRY;
	srb.HostAdapterID = adapter;
	SendASPI32Command(&srb);
	while(srb.CommandStatus==SS_PENDING) ;
	if (srb.CommandStatus!=SS_COMPLETE) return 0;
	if (srb.Parameters[3]==0) srb.Parameters[3] = 8;
	return ( (((int)srb.Parameters[3])<<8)|srb.HostAdapterSCSIID );
}

BOOL Inquiry( int adapter, int target, int lun, unsigned char* lpInqBuf )
{
	SRB_ExecIO	srb;
	ZeroMemory( &srb, sizeof(srb) );
	srb.HostAdapterID = adapter;
	srb.TargetID = target;
	srb.TargetLUN = lun;
	srb.DataAllocationLength = 32;
	srb.SenseAllocationLength = 14;
	srb.DataBufferPointer = lpInqBuf;
	srb.CDBLength = 6;
	srb.CDB[0] = SCSI_Inquiry;
	srb.CDB[4] = 32;
	if ( PerformAspiCmd( &srb ) )
		if (srb.CommandStatus==SS_COMPLETE) return TRUE;
	return FALSE;
}

BOOL TestUnitReady( int adapter, int target, int lun )
{
	SRB_ExecIO	srb;
	ZeroMemory( &srb, sizeof(srb) );
	unsigned char RsltBuf[MAX_LEN];
	srb.HostAdapterID = adapter;
	srb.TargetID = target;
	srb.TargetLUN = lun;
	srb.SenseAllocationLength = 16;
	srb.DataBufferPointer = RsltBuf;
	srb.CDBLength = 6;
	srb.CDB[0] = SCSI_TestUnitReady;
	return PerformAspiCmd( &srb );
}

BOOL StartStopUnit( int adapter, int target, int lun, int flag )
{
SRB_ExecIO*	psrb;
unsigned char* lpRsltBuf;
static DWORD dwThreadID;

	psrb = (SRB_ExecIO*)GlobalAlloc( GPTR, sizeof(SRB_ExecIO) );
	lpRsltBuf = (unsigned char*)GlobalAlloc( GMEM_FIXED, MAX_LEN );
	psrb->HostAdapterID = adapter;
	psrb->TargetID = target;
	psrb->TargetLUN = lun;
	psrb->SenseAllocationLength = 14;
	psrb->DataBufferPointer = lpRsltBuf;
	psrb->CDBLength = 6;
	psrb->CDB[0] = SCSI_StartStopUnit;
	psrb->CDB[4] = flag;			// 2 = eject, 3 = mount

	AddListItem(
		CreateThread( NULL, 0,
			(LPTHREAD_START_ROUTINE)PerformAspiCmdThread, (LPVOID)psrb,
			0, &dwThreadID )
		);
	return TRUE;
}


// TuXbhǗ֐

struct ListItem;
struct ListItem {
	HANDLE handle;
	ListItem* pnext;
};

ListItem* ptopitem = NULL;
ListItem** plastitempointer = &ptopitem;

// XbhXgŜāA
// IXbh΃nhăXg폜
int OptimizeList()
{
	int count = 0;
	DWORD dwExitCode;
	ListItem* pitem = ptopitem;
	ListItem** pprevitempointer = &ptopitem;
	while(pitem) {
		GetExitCodeThread( pitem->handle, &dwExitCode );
		if (dwExitCode == STILL_ACTIVE) {
			count++;
			pprevitempointer = &(pitem->pnext);
			pitem = pitem->pnext;
		} else {
			CloseHandle( pitem->handle );
			*pprevitempointer = pitem->pnext;
			delete pitem;
			pitem = *pprevitempointer;
		}
	}
	plastitempointer = pprevitempointer;
	return count;
}

// XbhXgɒǉ
int AddListItem( HANDLE hThread )
{
	// XbhXg̖ɃXbhnhǉ
	ListItem* pitem = new ListItem;
	pitem->handle = hThread;
	pitem->pnext = NULL;
	*plastitempointer = pitem;
	plastitempointer = &(pitem->pnext);
	// XbhXgŜďIXbh̓nhăXg폜
	return OptimizeList();
}

// XbhXg̑SẴXbhI܂ő҂
BOOL WaitAllThread( DWORD dwTimeout )
{
	int count = OptimizeList();
	HANDLE* ptable = new HANDLE[count];
	ListItem* pitem = ptopitem;
	count = 0;
	while(pitem) {
		ptable[count] = pitem->handle;
		count++;
		pitem = pitem->pnext;
	}
	DWORD ret = WaitForMultipleObjects(count, ptable, TRUE, dwTimeout);
	delete[] ptable;
	OptimizeList();
	return (ret!=WAIT_FAILED && ret!=WAIT_TIMEOUT);
	/* - pending method
	int count;
	do {
		count = OptimizeList();
	} while(count>0);
	return TRUE;
	*/
}


// ዉ֐

BOOL PerformAspiCmd( SRB_ExecIO* psrb )
{
HANDLE hASPICompleationEvent;
DWORD ASPIEventStatus;

	psrb->CommandCode = SC_EXEC_SCSI_CMD;
	psrb->RequestFlags = SRB_DIR_IN;

	hASPICompleationEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
	if ( hASPICompleationEvent == NULL ) return PerformAspiCmdPending( psrb );

	psrb->RequestFlags |= SRB_EVENT_NOTIFY;
	psrb->PostProc = hASPICompleationEvent;

	SendASPI32Command( psrb );

	if ( psrb->CommandStatus == SS_PENDING )
		ASPIEventStatus = WaitForSingleObject( hASPICompleationEvent, 10*1000 );

	CloseHandle(hASPICompleationEvent);

	return (psrb->CommandStatus==SS_COMPLETE);
}

BOOL PerformAspiCmdPending( SRB_ExecIO* psrb )
{
	SendASPI32Command(psrb);
	while(psrb->CommandStatus==SS_PENDING) ;
	return (psrb->CommandStatus==SS_COMPLETE);
}

BOOL PerformAspiCmdThread( SRB_ExecIO* psrb )
{
HANDLE hASPICompleationEvent;
DWORD ASPIEventStatus;

	psrb->CommandCode = SC_EXEC_SCSI_CMD;
	psrb->RequestFlags = SRB_DIR_IN;

	hASPICompleationEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
	if ( hASPICompleationEvent == NULL ) {
		ExitThread(1);
		return 1;
	}

	psrb->RequestFlags |= SRB_EVENT_NOTIFY;
	psrb->PostProc = hASPICompleationEvent;

	SendASPI32Command( psrb );

	if ( psrb->CommandStatus == SS_PENDING )
		ASPIEventStatus = WaitForSingleObject( hASPICompleationEvent, 10*1000 );

	CloseHandle(hASPICompleationEvent);

	GlobalFree(psrb->DataBufferPointer);
	GlobalFree(psrb);

	ExitThread(0);
	return 0;
}

////////////////////////////////////////
// ASPI32.CPP  Copyright(C) KEY(Yuu Igarashi) 1997,1998
